@@ -1,10 +1,30 @@
+mod search_table;
+
 use super::{State, StateIdx, DFA};
+use search_table::SearchTableSet;
+
 use crate::ast::{RuleKind, RuleRhs};
+use crate::range_map::RangeMap;
+
+use std::convert::TryFrom;
 
 use fxhash::FxHashMap;
-use proc_macro2::TokenStream;
+use proc_macro2::{Span, TokenStream};
 use quote::{quote, ToTokens};
 
+// Max. size for guards in ranges. When a case have more ranges than this we generate a binary
+// search table.
+//
+// Using binary search for large number of guards should be more efficient in runtime, but more
+// importantly, when using builtin regexes like `$$uppercase` that has a lot of cases (see
+// `char_ranges` module), rustc uses GiBs of RAM when compiling the generated code, even in debug
+// mode. For example, the test `builtins` takes more than 32GiB of memory to compile.
+//
+// Binary search does less comparisons in the worst case when we have more than 3 cases, but the
+// code for binary search is more complicated than a chain of `||`s, so I think it makes sense to
+// have a slightly larger number here.
+const MAX_GUARD_SIZE: usize = 9;
+
 pub fn reify(
     dfa: &DFA<Option<RuleRhs>>,
     user_state_type: Option<syn::Type>,
@@ -23,18 +43,21 @@ pub fn reify(
 
     let handle_type_name = syn::Ident::new(&(type_name.to_string() + "Handle"), type_name.span());
 
+    let mut search_tables = SearchTableSet::new();
+
     let match_arms = generate_state_arms(
         dfa,
         &handle_type_name,
         &action_enum_name,
         user_error_type.as_ref(),
         &token_type,
+        &mut search_tables,
     );
 
     let rule_name_enum_name = syn::Ident::new(&(type_name.to_string() + "Rule"), type_name.span());
     let rule_name_idents: Vec<syn::Ident> = rule_states
         .keys()
-        .map(|rule_name| syn::Ident::new(rule_name, proc_macro2::Span::call_site()))
+        .map(|rule_name| syn::Ident::new(rule_name, Span::call_site()))
         .collect();
 
     let switch_method = generate_switch(&rule_name_enum_name, rule_states);
@@ -43,7 +66,7 @@ pub fn reify(
 
     let lexer_error_type = match user_error_type {
         None => quote!(
-            #[derive(Debug, PartialEq, Eq)]
+            #[derive(Debug, PartialEq, Eq, Clone, Copy)]
             #visibility struct LexerError {
                 char_idx: usize,
             }
@@ -57,6 +80,44 @@ pub fn reify(
         ),
     };
 
+    let binary_search_fn = if search_tables.is_empty() {
+        quote!()
+    } else {
+        quote!(
+            fn binary_search(c: char, table: &[(char, char)]) -> bool {
+                table
+                    .binary_search_by(|(start, end)| match c.cmp(start) {
+                        std::cmp::Ordering::Greater => {
+                            if c <= *end {
+                                std::cmp::Ordering::Equal
+                            } else {
+                                std::cmp::Ordering::Less
+                            }
+                        }
+                        std::cmp::Ordering::Equal => std::cmp::Ordering::Equal,
+                        std::cmp::Ordering::Less => std::cmp::Ordering::Greater,
+                    })
+                    .is_ok()
+            }
+        )
+    };
+
+    let search_tables: Vec<TokenStream> = search_tables
+        .iter()
+        .map(|(ranges, ident)| {
+            let n_ranges = ranges.len();
+            let pairs: Vec<TokenStream> = ranges
+                .iter()
+                .map(|(start, end)| quote!((#start, #end)))
+                .collect();
+            quote!(
+                static #ident: [(char, char); #n_ranges] = [
+                    #(#pairs),*
+                ];
+            )
+        })
+        .collect();
+
     quote!(
         // Possible outcomes of a user action
         enum #action_enum_name<T> {
@@ -150,6 +211,9 @@ pub fn reify(
             #switch_method
         }
 
+        #(#search_tables)*
+        #binary_search_fn
+
         impl<'input> Iterator for #type_name<'input> {
             type Item = Result<(usize, #token_type, usize), LexerError<#(#user_error_type_lifetimes),*>>;
 
@@ -172,7 +236,7 @@ fn generate_switch(
     let mut arms: Vec<TokenStream> = vec![];
 
     for (rule_name, StateIdx(state_idx)) in rule_states.iter() {
-        let rule_ident = syn::Ident::new(rule_name, proc_macro2::Span::call_site());
+        let rule_ident = syn::Ident::new(rule_name, Span::call_site());
         arms.push(quote!(
             #enum_name::#rule_ident =>
                 self.state = #state_idx
@@ -196,6 +260,7 @@ fn generate_state_arms(
     action_enum_name: &syn::Ident,
     user_error_type: Option<&syn::Type>,
     token_type: &syn::Type,
+    search_tables: &mut SearchTableSet,
 ) -> Vec<TokenStream> {
     let DFA { states } = dfa;
 
@@ -223,8 +288,8 @@ fn generate_state_arms(
             assert!(*initial);
 
             // Initial state. Difference from other states is we return `None` when the
-            // iterator ends. In non-initial states EOF returns the last (longest) match, or
-            // fails (error).
+            // stream ends. In non-initial states EOS (end-of-stream) returns the last (longest)
+            // match, or fails (error).
             let error = make_lexer_error(quote!(self.current_match_start));
             let action = quote!({
                 return Some(Err(#error));
@@ -236,6 +301,7 @@ fn generate_state_arms(
                 range_transitions,
                 fail_transition,
                 &action,
+                search_tables,
             );
 
             quote!(
@@ -270,8 +336,9 @@ fn generate_state_arms(
                 *initial,
                 char_transitions,
                 range_transitions,
-                fail_transition,
+                &None,
                 &action,
+                search_tables,
             );
 
             if char_transitions.is_empty() && range_transitions.is_empty() {
@@ -294,9 +361,14 @@ fn generate_state_arms(
             // Non-initial, non-accepting state. In a non-accepting state we want to consume a
             // character anyway so we can use `next` instead of `peek`.
             let error = make_lexer_error(quote!(self.current_match_start));
-            let action = quote!({
-                return Some(Err(#error));
-            });
+            let action = match fail_transition {
+                None => quote!({
+                    return Some(Err(#error));
+                }),
+                Some(StateIdx(fail_state)) => quote!({
+                    self.state = #fail_state;
+                }),
+            };
 
             let state_char_arms = generate_state_char_arms(
                 *initial,
@@ -304,10 +376,22 @@ fn generate_state_arms(
                 range_transitions,
                 fail_transition,
                 &action,
+                search_tables,
             );
 
+            let end_of_stream_action = if *initial {
+                // In an initial state other than the state 0 we fail with "unexpected EOF"
+                quote!({
+                    return Some(Err(#error));
+                })
+            } else {
+                // Otherwise we run the semantic action and go to initial state of the current DFA.
+                // Initial state will then fail.
+                action
+            };
+
             quote!(match self.iter.peek().copied() {
-                None => return Some(Err(#error)),
+                None => #end_of_stream_action,
                 Some((char_idx, char)) => {
                     match char {
                         #(#state_char_arms,)*
@@ -316,24 +400,28 @@ fn generate_state_arms(
             })
         };
 
+        let state_idx_code = if state_idx == states.len() - 1 {
+            quote!(_)
+        } else {
+            quote!(#state_idx)
+        };
+
         match_arms.push(quote!(
-            #state_idx => #state_code
+            #state_idx_code => #state_code
         ));
     }
 
-    match_arms.push(quote!(_ => unreachable!()));
-
     match_arms
 }
 
-/// Generate arms on `match self.iter.next() { ... }` (for initial state) or `match
-/// self.iter.peek().copied() { ... }` (for other states) of DFA state.
+/// Generate arms for `match self.iter.peek().copied() { ... }`
 fn generate_state_char_arms(
     initial: bool,
     char_transitions: &FxHashMap<char, StateIdx>,
-    range_transitions: &FxHashMap<(char, char), StateIdx>,
+    range_transitions: &RangeMap<StateIdx>,
     fail_transition: &Option<StateIdx>,
     action: &TokenStream,
+    search_tables: &mut SearchTableSet,
 ) -> Vec<TokenStream> {
     // Arms of the `match` for the current character
     let mut state_char_arms: Vec<TokenStream> = vec![];
@@ -359,21 +447,26 @@ fn generate_state_char_arms(
 
     // Add range transitions. Same as above, use chain of "or"s for ranges with same transition.
     let mut state_ranges: FxHashMap<StateIdx, Vec<(char, char)>> = Default::default();
-    for (range, state_idx) in range_transitions {
-        state_ranges.entry(*state_idx).or_default().push(*range);
+    for range in range_transitions.iter() {
+        assert_eq!(range.values.len(), 1);
+        state_ranges.entry(range.values[0]).or_default().push((
+            char::try_from(range.start).unwrap(),
+            char::try_from(range.end).unwrap(),
+        ));
     }
 
-    for (StateIdx(next_state), mut ranges) in state_ranges.into_iter() {
-        let guard = if ranges.len() == 1 {
-            let (range_begin, range_end) = ranges.pop().unwrap();
-            quote!(x >= #range_begin && x <= #range_end)
+    for (StateIdx(next_state), ranges) in state_ranges.into_iter() {
+        let guard = if ranges.len() > MAX_GUARD_SIZE {
+            let binary_search_table_id = search_tables.add_table(ranges);
+
+            quote!(binary_search(x, &#binary_search_table_id))
         } else {
-            let (range_begin, range_end) = ranges.pop().unwrap();
-            let mut guard = quote!(x >= #range_begin && x <= #range_end);
-            while let Some((range_begin, range_end)) = ranges.pop() {
-                guard = quote!((x >= #range_begin && x <= #range_end) || #guard);
-            }
-            guard
+            let range_checks: Vec<TokenStream> = ranges
+                .into_iter()
+                .map(|(range_begin, range_end)| quote!((x >= #range_begin && x <= #range_end)))
+                .collect();
+
+            quote!(#(#range_checks)||*)
         };
 
         state_char_arms.push(quote!(
